@@ -9,6 +9,14 @@
 #include <string.h>
 
 
+#if defined(__clang__)
+# pragma clang diagnostic push
+# pragma clang diagnostic ignored "-Wdocumentation"
+# pragma clang diagnostic ignored "-Wunknown-attributes"
+#endif
+
+
+
 /**
  * Only include some C code if compiling with GCC.
  * 
@@ -280,7 +288,7 @@ struct libkeccak_state {
  * @param  x     The value of x in `SHA3-x`, the output size
  */
 LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__)))
-static inline void
+inline void
 libkeccak_spec_sha3(struct libkeccak_spec *restrict spec, long int x)
 {
 	spec->bitrate = 1600 - 2 * x;
@@ -296,7 +304,7 @@ libkeccak_spec_sha3(struct libkeccak_spec *restrict spec, long int x)
  * @param  d     The output size
  */
 LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__)))
-static inline void
+inline void
 libkeccak_spec_rawshake(struct libkeccak_spec *restrict spec, long int x, long int d)
 {
 	spec->bitrate = 1600 - 2 * x;
@@ -319,8 +327,8 @@ libkeccak_spec_rawshake(struct libkeccak_spec *restrict spec, long int x, long i
  * @param   spec  The specifications datastructure to check
  * @return        Zero if error free, a `LIBKECCAK_SPEC_ERROR_*` if an error was found
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__, __unused__, __warn_unused_result__, __pure__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__, __warn_unused_result__, __pure__)))
+inline int
 libkeccak_spec_check(const struct libkeccak_spec *restrict spec)
 {
 	long int state_size = spec->capacity + spec->bitrate;
@@ -350,8 +358,8 @@ libkeccak_spec_check(const struct libkeccak_spec *restrict spec)
  * 
  * @param  spec  The specification datastructure to fill in
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__, __unused__)))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__)))
+inline void
 libkeccak_generalised_spec_initialise(struct libkeccak_generalised_spec *restrict spec)
 {
 	spec->bitrate    = LIBKECCAK_GENERALISED_SPEC_AUTOMATIC;
@@ -386,8 +394,8 @@ int libkeccak_state_initialise(struct libkeccak_state *restrict, const struct li
  * 
  * @param  state  The state that should be reset
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__, __unused__)))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__)))
+inline void
 libkeccak_state_reset(struct libkeccak_state *restrict state)
 {
 	state->mptr = 0;
@@ -399,7 +407,7 @@ libkeccak_state_reset(struct libkeccak_state *restrict state)
  * 
  * @param  state  The state that should be destroyed
  */
-static inline void
+inline void
 libkeccak_state_fast_destroy(struct libkeccak_state *restrict state)
 {
 	if (state) {
@@ -437,8 +445,8 @@ void libkeccak_state_wipe(volatile struct libkeccak_state *restrict);
  * 
  * @param  state  The state that should be destroyed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__, __optimize__("-O0"))))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__optimize__("-O0"))))
+inline void
 libkeccak_state_destroy(volatile struct libkeccak_state *restrict state)
 {
 	if (state) {
@@ -454,25 +462,15 @@ libkeccak_state_destroy(volatile struct libkeccak_state *restrict state)
  * @param   spec  The specifications for the state
  * @return        The state, `NULL` on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __unused__, __warn_unused_result__, __malloc__)))
-static inline struct libkeccak_state *
-libkeccak_state_create(const struct libkeccak_spec *restrict spec)
-{
-	struct libkeccak_state *restrict state = malloc(sizeof(struct libkeccak_state));
-	if (!state || libkeccak_state_initialise(state, spec)) {
-		free(state);
-		return NULL;
-	}
-	return state;
-}
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __warn_unused_result__, __malloc__)))
+struct libkeccak_state *libkeccak_state_create(const struct libkeccak_spec *restrict);
 
 /**
  * Wrapper for `libkeccak_state_fast_destroy` that also frees the allocation of the state
  * 
  * @param  state  The state that should be freed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__)))
-static inline void
+inline void
 libkeccak_state_fast_free(struct libkeccak_state *restrict state)
 {
 	libkeccak_state_fast_destroy(state);
@@ -484,8 +482,8 @@ libkeccak_state_fast_free(struct libkeccak_state *restrict state)
  * 
  * @param  state  The state that should be freed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__, __optimize__("-O0"))))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__optimize__("-O0"))))
+inline void
 libkeccak_state_free(volatile struct libkeccak_state *restrict state)
 {
 #ifdef __GNUC__
@@ -515,17 +513,8 @@ int libkeccak_state_copy(struct libkeccak_state *restrict, const struct libkecca
  * @param   src  The state to duplicate
  * @return       The duplicate, `NULL` on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __unused__, __warn_unused_result__, __malloc__)))
-static inline struct libkeccak_state *
-libkeccak_state_duplicate(const struct libkeccak_state *restrict src)
-{
-	struct libkeccak_state *restrict dest = malloc(sizeof(struct libkeccak_state));
-	if (!dest || libkeccak_state_copy(dest, src)) {
-		libkeccak_state_free(dest);
-		return NULL;
-	}
-	return dest;
-}
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __warn_unused_result__, __malloc__)))
+struct libkeccak_state *libkeccak_state_duplicate(const struct libkeccak_state *restrict);
 
 /**
  * Marshal a `struct libkeccak_state` into a buffer
@@ -548,6 +537,38 @@ size_t libkeccak_state_marshal(const struct libkeccak_state *restrict, void *res
 LIBKECCAK_GCC_ONLY(__attribute__((__leaf__, __nonnull__(2))))
 size_t libkeccak_state_unmarshal(struct libkeccak_state *restrict, const void *restrict);
 
+/**
+ * Get the number of bytes that are absorbed during
+ * one pass of the absorption phase
+ * 
+ * @param   state  The hashing state
+ * @return         The number of bytes absorbed during one pass
+ */
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__, __warn_unused_result__, __pure__)))
+inline size_t
+libkeccak_zerocopy_chunksize(struct libkeccak_state *restrict state)
+{
+	return state->r >> 3;
+}
+
+/**
+ * Absorb more of the message to the Keccak sponge
+ * without copying the data to an internal buffer
+ * 
+ * It is safe run zero-copy functions before non-zero-copy
+ * functions for the same state, running zero-copy functions
+ * after non-zero-copy functions on the other hand can
+ * cause the message to be misread
+ * 
+ * @param  state   The hashing state
+ * @param  msg     The partial message
+ * @param  msglen  The length of the partial message; must be a
+ *                 multiple of `libkeccak_zerocopy_chunksize(state)`
+ *                 (undefined behaviour otherwise)
+ */
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __nothrow__)))
+void libkeccak_zerocopy_update(struct libkeccak_state *restrict, const void *restrict, size_t);
+
 /**
  * Absorb more of the message to the Keccak sponge
  * without wiping sensitive data when possible
@@ -572,6 +593,30 @@ int libkeccak_fast_update(struct libkeccak_state *restrict, const void *restrict
 LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__)))
 int libkeccak_update(struct libkeccak_state *restrict, const void *restrict, size_t);
 
+
+/**
+ * Absorb the last part of the message and squeeze the Keccak sponge
+ * without copying the data to an internal buffer
+ * 
+ * It is safe run zero-copy functions before non-zero-copy
+ * functions for the same state, running zero-copy functions
+ * after non-zero-copy functions on the other hand can
+ * cause the message to be misread
+ * 
+ * @param  state    The hashing state
+ * @param  msg      The rest of the message; will be edited; extra memory
+ *                  shall be allocated such that `suffix` and a 10*1 pad (which
+ *                  is at least 2 bits long) can be added in a why the makes it's
+ *                  length a multiple of `libkeccak_zerocopy_chunksize(state)`
+ * @param  msglen   The length of the partial message
+ * @param  bits     The number of bits at the end of the message not covered by `msglen`
+ * @param  suffix   The suffix concatenate to the message, only '1':s and '0':s, and NUL-termination
+ * @param  hashsum  Output parameter for the hashsum, may be `NULL`
+ */
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(1, 2))))
+void libkeccak_zerocopy_digest(struct libkeccak_state *restrict, void *restrict, size_t,
+                               size_t, const char *restrict, void *restrict);
+
 /**
  * Absorb the last part of the message and squeeze the Keccak sponge
  * without wiping sensitive data when possible
@@ -688,8 +733,8 @@ int libkeccak_generalised_sum_fd(int, struct libkeccak_state *restrict, const st
  *                   at least `((spec->output + 7) / 8) * sizeof(char)`, may be `NULL`
  * @return           Zero on success, -1 on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2, 3), __artificial__, __gnu_inline__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2, 3), __artificial__)))
+inline int
 libkeccak_keccaksum_fd(int fd, struct libkeccak_state *restrict state,
                        const struct libkeccak_spec *restrict spec, void *restrict hashsum)
 {
@@ -707,8 +752,8 @@ libkeccak_keccaksum_fd(int fd, struct libkeccak_state *restrict state,
  *                   at least `((output + 7) / 8) * sizeof(char)`, may be `NULL`
  * @return           Zero on success, -1 on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__, __gnu_inline__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__)))
+inline int
 libkeccak_sha3sum_fd(int fd, struct libkeccak_state *restrict state, long output, void *restrict hashsum)
 {
 	struct libkeccak_spec spec;
@@ -728,8 +773,8 @@ libkeccak_sha3sum_fd(int fd, struct libkeccak_state *restrict state, long output
  *                        at least `((output + 7) / 8) * sizeof(char)`, may be `NULL`
  * @return                Zero on success, -1 on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__, __gnu_inline__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__)))
+inline int
 libkeccak_rawshakesum_fd(int fd, struct libkeccak_state *restrict state, long semicapacity, long output, void *restrict hashsum)
 {
 	struct libkeccak_spec spec;
@@ -749,8 +794,8 @@ libkeccak_rawshakesum_fd(int fd, struct libkeccak_state *restrict state, long se
  *                        at least `((output + 7) / 8) * sizeof(char)`, may be `NULL`
  * @return                Zero on success, -1 on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__, __gnu_inline__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(2), __artificial__)))
+inline int
 libkeccak_shakesum_fd(int fd, struct libkeccak_state *restrict state, long semicapacity, long output, void *restrict hashsum)
 {
 	struct libkeccak_spec spec;
@@ -808,7 +853,7 @@ struct libkeccak_hmac_state {
 	 */
 	unsigned char leftover;
 
-	char __pad[sizeof(void *) / sizeof(char) - 1];
+	char __pad[sizeof(void *) - 1];
 };
 
 
@@ -833,7 +878,7 @@ int libkeccak_hmac_set_key(struct libkeccak_hmac_state *restrict, const void *re
  * @return              Zero on success, -1 on error
  */
 LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__)))
-static inline int
+inline int
 libkeccak_hmac_initialise(struct libkeccak_hmac_state *restrict state, const struct libkeccak_spec *restrict spec,
                           const void *restrict key, size_t key_length)
 {
@@ -857,8 +902,8 @@ libkeccak_hmac_initialise(struct libkeccak_hmac_state *restrict state, const str
  * @param   key_length  The length of key, in bits
  * @return              The state, `NULL` on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __unused__, __warn_unused_result__, __malloc__)))
-static inline struct libkeccak_hmac_state *
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __warn_unused_result__, __malloc__)))
+inline struct libkeccak_hmac_state *
 libkeccak_hmac_create(const struct libkeccak_spec *restrict spec, const void *restrict key, size_t key_length)
 {
 	struct libkeccak_hmac_state *restrict state = malloc(sizeof(struct libkeccak_hmac_state));
@@ -878,8 +923,8 @@ libkeccak_hmac_create(const struct libkeccak_spec *restrict spec, const void *re
  * @param   key_length  The length of key, in bits, ignored if `key == NULL`
  * @return              Zero on success, -1 on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(1), __unused__)))
-static inline int
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(1))))
+inline int
 libkeccak_hmac_reset(struct libkeccak_hmac_state *restrict state, const void *restrict key, size_t key_length)
 {
 	libkeccak_state_reset(&state->sponge);
@@ -899,7 +944,7 @@ void libkeccak_hmac_wipe(volatile struct libkeccak_hmac_state *restrict);
  * 
  * @param  state  The state that should be destroyed
  */
-static inline void
+inline void
 libkeccak_hmac_fast_destroy(struct libkeccak_hmac_state *restrict state)
 {
 	if (!state)
@@ -918,8 +963,8 @@ libkeccak_hmac_fast_destroy(struct libkeccak_hmac_state *restrict state)
  * 
  * @param  state  The state that should be destroyed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__, __optimize__("-O0"))))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__optimize__("-O0"))))
+inline void
 libkeccak_hmac_destroy(volatile struct libkeccak_hmac_state *restrict state)
 {
 	if (!state)
@@ -940,8 +985,7 @@ libkeccak_hmac_destroy(volatile struct libkeccak_hmac_state *restrict state)
  * 
  * @param  state  The state that should be freed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__)))
-static inline void
+inline void
 libkeccak_hmac_fast_free(struct libkeccak_hmac_state *restrict state)
 {
 	libkeccak_hmac_fast_destroy(state);
@@ -953,8 +997,8 @@ libkeccak_hmac_fast_free(struct libkeccak_hmac_state *restrict state)
  * 
  * @param  state  The state that should be freed
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__unused__, __optimize__("-O0"))))
-static inline void
+LIBKECCAK_GCC_ONLY(__attribute__((__optimize__("-O0"))))
+inline void
 libkeccak_hmac_free(volatile struct libkeccak_hmac_state *restrict state)
 {
 #ifdef __GNUC__
@@ -984,13 +1028,15 @@ int libkeccak_hmac_copy(struct libkeccak_hmac_state *restrict, const struct libk
  * @param   src  The state to duplicate
  * @return       The duplicate, `NULL` on error
  */
-LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __unused__, __warn_unused_result__, __malloc__)))
-static inline struct libkeccak_hmac_state *
+LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__, __warn_unused_result__, __malloc__)))
+inline struct libkeccak_hmac_state *
 libkeccak_hmac_duplicate(const struct libkeccak_hmac_state *restrict src)
 {
-	struct libkeccak_hmac_state* restrict dest = malloc(sizeof(struct libkeccak_hmac_state));
-	if (!dest || libkeccak_hmac_copy(dest, src))
-		return libkeccak_hmac_free(dest), NULL;
+	struct libkeccak_hmac_state *restrict dest = malloc(sizeof(struct libkeccak_hmac_state));
+	if (!dest || libkeccak_hmac_copy(dest, src)) {
+		libkeccak_hmac_free(dest);
+		return NULL;
+	}
 	return dest;
 }
 
@@ -1002,17 +1048,24 @@ libkeccak_hmac_duplicate(const struct libkeccak_hmac_state *restrict src)
  * @return         The number of bytes stored to `data`
  */
 LIBKECCAK_GCC_ONLY(__attribute__((__nonnull__(1), __nothrow__)))
-static inline size_t
+inline size_t
 libkeccak_hmac_marshal(const struct libkeccak_hmac_state *restrict state, void *restrict data_)
 {
 	unsigned char *restrict data = data_;
-	size_t written = libkeccak_state_marshal(state ? &state->sponge : NULL, data);
+	size_t written = libkeccak_state_marshal(&state->sponge, data);
 	if (data) {
-		data += written / sizeof(char);
+		data += written;
+#if defined(__clang__)
+# pragma clang diagnostic push
+# pragma clang diagnostic ignored "-Wcast-align"
+#endif
 		*(size_t *)data = state->key_length;
-		data += sizeof(size_t) / sizeof(char);
+#if defined(__clang__)
+# pragma clang diagnostic pop
+#endif
+		data += sizeof(size_t);
 		memcpy(data, state->key_opad, (state->key_length + 7) >> 3);
-		data += ((state->key_length + 7) >> 3) / sizeof(char);
+		data += (state->key_length + 7) >> 3;
 		data[0] = (unsigned char)!!state->key_ipad;
 		data[1] = state->leftover;
 	}
@@ -1094,4 +1147,8 @@ int libkeccak_hmac_digest(struct libkeccak_hmac_state *restrict state, const voi
 #include "libkeccak-legacy.h"
 
 
+#if defined(__clang__)
+# pragma clang diagnostic pop
+#endif
+
 #endif
